// This is an implementation of an unbounded SystemVerilog
// mailbox that can contain a any kind of datatype;
// mailbox entries can even have different types. (See note
// below for explanation how to use this.) 
// This code was tested on Early Adopter 8 (v121).
//
// The following mailbox member tasks/functions are
// implemented:
//
//   - num
//   - put
//   - try_put
//   - get
//   - try_get
//   - peek
//   - try_peek
//
//
// The remaining method, try_put, only makes sense for
// unbounded mailboxes.
//
// To use this mailbox, change the 'mbox_data' typedef
// to define the type of data you want to store in the
// mailbox. Then import the mailbox into a program or
// module, and allocate an instance with new. For example:
//
// import mbox::*;
// mailbox MsgStack = new;
// 
// Declare the mailbox message using the 'mbox_data' 
// datatype:
//
// mbox_data mail;
// MsgStack.put(mail);
// MsgStack.get(mail);
//
// Notes on extending this:
//
// This is a complete version of a mailbox. It supports all kind of data types. 
// Any item which should be put in the mail box must be derived from the
// virtual class mb_item.

package mbox;
// This is the prototype for mailbox data itemChange this typedef to match the data you want to put
// in the mailbox.
// List item placed in mailbox.
virtual class mb_item;
   mb_item next;
   function new;
      next = null;
   endfunction
endclass

// Unbounded mailbox
class mailbox;
   // Size of mailbox
   local int count;
   // FIFO list: first, last ptr
   local mb_item first;
   local mb_item last;
   
   local int max_count;
   

   local bit debug_on = 0;
   local int mb_id;

   // Constructor   
   function new(int mc=0);
      count = 0;
      max_count = mc;
      first = null;
      last = null;
   endfunction

   // Returns number of available slots in the mailbox
   function int num();
      return(count);
   endfunction

   // Add an item to the mailbox -> Blocking.
   // If there is no space available wait until 
   // items have been taken of the mailbox from
   // the other side (->get() ) and new space 
   // becomes available again.
   task put(mb_item d);
      if ( (debug_on==1) && (count >= max_count) && (max_count!=0)) $display("MailBox %0d: Blocking PUT action until mailbox is not full.", mb_id);
      while ((count >= max_count) && (max_count!=0)) @count;
      if (first == null)
         first = d;
      else
         last.next = d;
      last = d;
      count++;
      if (debug_on==1) begin
         $display("MailBox %0d: t(PUT) mb_item NEW_PUT %0d: ", mb_id, d);
         $display("MailBox %0d: t(PUT) mb_item FIRST   %0d: ", mb_id, first);
         $display("MailBox %0d: t(PUT) mb_item LAST    %0d: ", mb_id, last);
         $display("MailBox %0d: t(PUT) mb_item COUNT   %0d: ", mb_id, count);
      end	 
   endtask


   // Add an item to the mailbox, returning
   // immediately if there is no space left
   // -> Non-Blocking
   function int try_put(mb_item d);
      try_put = 0;
      if ((count < max_count) || (max_count==0)) begin
         if (first == null)
            first = d;
         else
            last.next = d;
         last = d;
         count++;
         if (debug_on==1) begin
            $display("MailBox %0d: t(PUT) mb_item NEW_PUT %0d: ", mb_id, d);
            $display("MailBox %0d: t(PUT) mb_item FIRST   %0d: ", mb_id, first);
            $display("MailBox %0d: t(PUT) mb_item LAST	  %0d: ", mb_id, last);
            $display("MailBox %0d: t(PUT) mb_item COUNT   %0d: ", mb_id, count);
         end
	 try_put = 1;
      end	 	    
   endfunction


   // Remove an item from the mailbox. Can
   // block if the mailbox is empty.
   task get( ref mb_item d);
      if ( (debug_on==1) && (count <= 0)) $display("MailBox %0d: Blocking GET action until mailbox is not empty.", mb_id);
      while (count <= 0) @count;
      d = pop();
      if (debug_on==1) begin
         $display("MailBox %0d: t(GET) mb_item NEW_GET %0d: ", mb_id, d);
      end
   endtask


   // Remove an item from the mailbox, returning
   // immediately if there is nothing. This method
   // cannot block. Returns 0 if the mailbox was
   // empty, 1 if the mailbox had data.
   function int try_get(mb_item d);
      try_get = 0;
      if (count > 0)
         begin
         d = pop();
         try_get = 1;
         end
   endfunction


   // Internal helper function to remove an item
   // from the mailbox
   local function mb_item pop();
      pop = first;
       if (first == last)
         begin
         first = null;
         last = null;
         count = 0;
         end
      else
         begin
         first = first.next;
         count--;
         end
      if (debug_on==1) begin
         $display("MailBox %0d: t(POP) mb_item NEW_POP %0d: ", mb_id, pop);
         $display("MailBox %0d: t(POP) mb_item FIRST   %0d: ", mb_id, first);
         $display("MailBox %0d: t(POP) mb_item LAST    %0d: ", mb_id, last);
         $display("MailBox %0d: t(POP) mb_item COUNT   %0d: ", mb_id, count);
      end	 
   endfunction
    

   // Get the first item in the mailbox, without
   // removing it from the mailbox. Blocks if the
   // mailbox is empty.
   task peek(mb_item d);
      while (count <= 0) @count;
      d = first;
   endtask

   // Get the first item in the mailbox, if it
   // exists. Does not block if the mailbox is
   // empty.
   function int try_peek(mb_item d);
      try_peek = 0;
      if (count > 0)
         begin
         d = first;
         try_peek = 1;
         end
   endfunction
   
   task mb_debug(bit onoff=0);
      debug_on = onoff;
   endtask
   
   task set_mb_id(int id);
      mb_id = id;
   endtask
   
   function int get_mb_id();
      get_mb_id = mb_id;
   endfunction

endclass

endpackage
